// ====================== elements/local/myFirstElement.cc ================
// 
// Written by: Alexander Jacobsen
// Element Description: Element which does nothing...
// Last Updated: 07.03.2011
//
// ========================================================================

// Mandatory inclusions. 
#include<click/config.h>
#include<click/confparse.hh>
#include<click/error.hh>
#include<click/string.hh>
#include<click/timestamp.hh>
#include <click/straccum.hh>
#include"MyFirstElement.hh"


//All click declarations are enclosed within a macro pair, CLICK_DECLS and CLICKENDDECLS.
CLICK_DECLS
	
	MyFirstElement::MyFirstElement() : _q(0)
	//MyFirstElement::MyFirstElement()
	{ 
		_totalPacketCounter = 0;
		_tempPacketCounter = 0;
		_capacity = 1000;
		oldTimeStamp.assign_now();
		_head = 0;
		_tail = 0;
		_drops = 0;
		//Set the leave time interval in ms.
		_leaveTime = 0;
		_networkDestAddrOffset = 16;
		
		//Nr of IPs that id a flow.
		_nrOfIps = 0;

		//Sets the ip destination as id for a flow. Used for id and manipulation of
		//flows.
		_destinationIPs = "158.38.122.8:50,158.38.122.10:50,192.145.123.3:60,122.4.2.2:100,";

		_annotationOffset = 16;

		//Offset which returns the source ip address
		_networkSourceAddrOffset = 12;

		packetLeaveTime = new Timestamp[_capacity];

		listOfDestinationIPsToManipulate = new String[1000];

		kpMetrics = new String[1000];



		_increaseAckDelay = true;

	}
	MyFirstElement::~MyFirstElement() 
	{ 
	
		
	}

	static String
	averagecounter_read_time_between_ACKs_handler(Element *e)
	{
		MyFirstElement *c = (MyFirstElement *)e;
	  	return String(c->averageTimeBetweenAcks());
	}

	int MyFirstElement::initialize(ErrorHandler *errh)
	{

		assert(!_q && _head == 0 && _tail == 0);
		//Array which stores pointers to each individual packet.
		_q = (Packet **) CLICK_LALLOC(sizeof(Packet *) * (_capacity + 1));
		


		if (_q == 0)
			return errh->error("out of memory");

		return 0;
	}

	//Add a handler which makes it possible to read the nr of packets.
	void MyFirstElement::add_handlers() {
    	add_data_handlers("nrofpackets", Handler::OP_READ | Handler::OP_WRITE, &_totalPacketCounter);
    	add_data_handlers("packetDelay", Handler::OP_READ | Handler::OP_WRITE, &_leaveTime);
    	add_data_handlers("destinationIPs", Handler::OP_READ | Handler::OP_WRITE, &_destinationIPs);
		add_read_handler("averageTimeBetweenAcks", averagecounter_read_time_between_ACKs_handler, &_averageTimeBetweenAcks);
		add_read_handler("length", read_handler, (void *)0);
	}

	//Defines the push method. This is where a packet arrives if an upstream element transfers the packet to
	//this element.
	void MyFirstElement::push(int port, Packet *p) {

		//Sets the new timestamp to current time since 1970.
		newTimeStamp.assign_now();

		//Counts the number of packets traversing the element. One for total amount
		//of packets, the other used to calculate average over the last 1000 packets.
		_totalPacketCounter++;
		_tempPacketCounter++;

		//Adds the time between this and the previous packet to the
		//total time between packets.
		totalTimeStampBetweenAcks += (newTimeStamp-oldTimeStamp);

		//Timestamp used to print the time between this and previous packet.
		tempTimeBetweenThisAndPreviousPack = (newTimeStamp-oldTimeStamp);

		//Calculates the average time between acks.
		_averageTimeBetweenAcks = calculateAverageBetweenACKs();

		//Reset average time between acks calc every 1000 pack.
		if( (_tempPacketCounter%1000) ==0)
		{
			_tempPacketCounter = 1;
			totalTimeStampBetweenAcks.assign(0);

		}

		sa << "************ Packet nr: " << String(_totalPacketCounter) << " ************" << "\n";

		//Prints the packet arrival time.
		sa << "Packet arrival time : " << newTimeStamp.sec() << " (sec) : " << newTimeStamp.msec()  << " (ms)\n";

		//Prints the difference between the arrival time of this and previous packet.
		sa << "Time between this and previous packet : " << tempTimeBetweenThisAndPreviousPack.unparse() << " (sec)\n";

		sa << "Total time between ACKs : " << totalTimeStampBetweenAcks.msec1() << " (ms)\n";

		//Prints the new timestamp in ms since 1970.		
		sa << "Average Time between the last 1000 ACKs : ";
		sa << calculateAverageBetweenACKs() << " (ms)\n";

		sa << "Source IP Address = ";
		sa << IPAddress(p->data() + _networkDestAddrOffset) << "\n";

		sa << "Destination IP Address = ";
		sa << IPAddress(p->data() + _networkSourceAddrOffset) << "\n";

		_packDestinationAddress = IPAddress(p->data() + _networkSourceAddrOffset).unparse();

		//Prints amount of dropped packets from queue.
		sa << "Number of packets dropped: " << _drops << "\n";

		//Print the destination addresses of packets that should be manipulated.
		sa << "IP dest addr of packs to manipulate flow rate : " << _destinationIPs << "\n";

		//Stor the IPs in a array.
		storeDestIPsToManipulate();

		//Assigns the newTimestamp to OldTimestamp for use as reference for next arriving packet.
		oldTimeStamp = newTimeStamp;

		//Enqueue packet.
		enqueue(p);

		//Write debug text to the click chatter.
		click_chatter(sa.c_str());
		sa.clear();
	}
	
	//Method which selects the next packet to send.
	//inline Packet *	MyFirstElement::getFirstPacketToSend()
	//{
	 //   int h = _head, t = _tail;

	   // timeCheck.assign_now();

	   // if (h != t && timeCheck.usec() >= packetLeaveTime[h] ) {
	    //	Packet *p = _q[h];
		//packet_memory_barrier(_q[h], _head);
		//_head = next_i(h);
		//assert(p);
		//return p;
	    //} else
		//return 0;
	//}

	inline bool	MyFirstElement::enqueue(Packet *p)
	{
		assert(p);
	    int h = _head, t = _tail;

	    //Check for the next free index.
	    //Returns t+1 if the next index is unoccupied.
	    //Returns 0 if t = capacity of the queue.
	    int nextUnoccupiedIndex = next_i(t);

	    //If next free position in the queue is not the head position, store
	    //the packet in the queue. Else, if the next free position is the head
	    //(start) of the queue, the queue is full and the packet is dropped.
	    if (nextUnoccupiedIndex != h) {

	    	//Put the packet in the position of the tail in the queue.
	    	//First time tail is 0, else it is the next unoccupied position.
	    	_q[t] = p;

	    	//Set packet nr, used for debugging.
	    	p->set_anno_s32( _annotationOffset, _totalPacketCounter );

	    	//Set the time when the packet should leave the element.
	    	setLeaveTimeForPacket(t);


	    	//packet_memory_barrier(_q[t], _tail);
	    	_tail = nextUnoccupiedIndex;

	    	//Gets the current size of the queue.
	    	//int s = size(h, nextUnoccupiedIndex);

	    	//if (s > _highwater_length)
	    		//_highwater_length = s;

	    	return true;
	    }
	    else {

	    	//Drop packet.
	    	p->kill();
	    	_drops++;

	    	return false;
	    }
	}

	//Function which sets the when the packet should leave the element.
	//The inline keyword is used for small functions to enhance performance.
	inline void	MyFirstElement::setLeaveTimeForPacket(int packetQueueIndex){
		Timestamp now;
		now.assign_now();

		bool calculateNewLeaveTime = false;

		//sa << "Time now (sec) : " << now.sec() << " (ms) : " << now.msec() << "\n";

		//Store the packet leave time in array. If packet should be delayed, new leave time is
		//calculated. Else the packet should leave ASAP.
		if( checkIfDestAddressShouldBeManipulated() )
		{
			packetLeaveTime[packetQueueIndex] = calculateLeaveTimeForPacket();
			sa << "Packet leave time :  " << packetLeaveTime[packetQueueIndex].sec() << " (sec) : " << packetLeaveTime[packetQueueIndex].msec() << " (ms)\n";
			sa << "Queue head : " << _head << "\n";
			sa << "Queue tail : " << _tail << "\n";
			//quicksort arrays of packets and leave times
			quickSort(_head,_tail);


		}
		else
		{
			packetLeaveTime[packetQueueIndex] = now;
			sa << "Packet leave time :  " << packetLeaveTime[packetQueueIndex].sec() << " (sec) : " << packetLeaveTime[packetQueueIndex].msec() << " (ms)\n";
			sa << "Queue head : " << _head << "\n";
			sa << "Queue tail : " << _tail << "\n";

			//quicksort arrays of packets and leave times
			quickSort(_head,_tail);
		}



	}

	//Function which checks if the destination of the address is recognized. If it is, then a new
	//leave time should be calculated. If not, set leavetime to now.
	inline bool MyFirstElement::checkIfDestAddressShouldBeManipulated(){
			bool packDestAddrRecognized = false;

			//Loop through all destination ips if they exist
			if(_nrOfIps!=0){

				for(int i = 0; i<_nrOfIps; i++)
				{
						if( listOfDestinationIPsToManipulate[i].compare(_packDestinationAddress) == 0 )
						{
							packDestAddrRecognized = true;
						}
				}
			}

			return packDestAddrRecognized;
	}


	//Function which calculates when the packet should leave the element.
	inline Timestamp MyFirstElement::calculateLeaveTimeForPacket(){
		//Sets the current time stamp.
		leaveTime.assign_now();

		//Calculates new time when the packet should leave.
		leaveTime = leaveTime.make_msec(leaveTime.sec() , ( leaveTime.msec()+_leaveTime));


		//if(_increaseAckDelay && _totalPacketCounter%100==0)
		//	_leaveTime+=1;
		//else if(_leaveTime==150){
		//	_increaseAckDelay = false;
			//_leaveTime = 0;
		//}

		return leaveTime;
	}

	//Function which clears the queues from memory.
	void MyFirstElement::cleanup(CleanupStage)
	{
	    for (int i = _head; i != _tail; i = next_i(i))
	    	_q[i]->kill();
	    CLICK_LFREE(_q, sizeof(Packet *) * (_capacity + 1));
	    _q = 0;

	    //Delete the array containing packets leave times.
	    delete [] packetLeaveTime;

	    //Make packetLeaveTime point to null
	    packetLeaveTime = 0;

	}
	//Method which is triggered when the element receives a request for a packet.
	Packet * MyFirstElement::pull(int)
	{
	    return deq();
	}

	String MyFirstElement::read_handler(Element *e, void *thunk)
	{
	    MyFirstElement *q = static_cast<MyFirstElement *>(e);
	    int which = reinterpret_cast<intptr_t>(thunk);
	    switch (which) {
	      case 0:
		return String(q->size());
	      default:
		return "";
	    }
	}
	//Calculates the average time between incoming ACKs
	int MyFirstElement::calculateAverageBetweenACKs(){

		//Adds the seconds from timestamp.
		int tempAvTimeBetweenAcksInMilliSecs = totalTimeStampBetweenAcks.msec1()/_tempPacketCounter;

		return tempAvTimeBetweenAcksInMilliSecs;
	}

	void MyFirstElement::storeDestIPsToManipulate(){

		int stringSize = _destinationIPs.length();
		int separatorPos = 0;
		int separatorPosIterator = 0;
		int colonSeparatorPos = 0;

		String rest = _destinationIPs;
		String tempIP = "";
		String tempKpMetric = "";

		_nrOfIps = 0;

		//Do as long as comma is found in string.
		while(separatorPos >= 0) {

			//Search for char in string. Returns -1 if the character is not found.
			colonSeparatorPos = rest.find_left(':' , 0);
			separatorPos = rest.find_left(',' , 0);

			//Do something if entries are found.
			if(separatorPos >= 0){

				//Update list of IPs
				tempIP = rest.substring(0,colonSeparatorPos);

				tempKpMetric = rest.substring(colonSeparatorPos+1, ( separatorPos-(colonSeparatorPos+1) ) );

				listOfDestinationIPsToManipulate[_nrOfIps] = tempIP;
				kpMetrics[_nrOfIps] = tempKpMetric;

				separatorPosIterator += separatorPos+1;

				rest = _destinationIPs.substring(separatorPosIterator,stringSize);

				stringSize = rest.length();

				_nrOfIps++;
			}
		}

		//Print the found entries for debugging.
		for(int i = 0; i<_nrOfIps; i++){

			sa << "IP : " << listOfDestinationIPsToManipulate[i] << " KP metric : " <<  kpMetrics[i] << "\n";

		}
	}

	void MyFirstElement::quickSort(int start, int end){

		int pivotPoint;

		if(start < end){

			//Get the pivot point
			pivotPoint = partition(start , end);

			//Sort the first sub list.
			quickSort(start,pivotPoint-1);

			//Sort the second sub list.
			quickSort(pivotPoint+1, end);
		}

	}

	int MyFirstElement::partition(int start, int end){

		Timestamp pivotValue;
		int pivotIndex, mid;

		mid = ( start+end )/2;
		swap(packetLeaveTime[start], packetLeaveTime[mid], _q[start], _q[mid]);

		pivotIndex = start;
		pivotValue = packetLeaveTime[start];

		for(int scan = start + 1; scan <= end; scan++){

			if(packetLeaveTime[scan] < pivotValue){

				pivotIndex++;
				swap(packetLeaveTime[pivotIndex], packetLeaveTime[scan], _q[pivotIndex], _q[scan]);

			}

		}
		swap(packetLeaveTime[start], packetLeaveTime[pivotIndex], _q[start], _q[pivotIndex]);
		return pivotIndex;
	}

	void MyFirstElement::swap(Timestamp &value1, Timestamp &value2, Packet *pack1, Packet *pack2){

		Timestamp temp = value1;
		Packet *tempPack = pack1;

		value1 = value2;
		pack1 = pack2;

		value2 = temp;
		pack2 = tempPack;
	}




CLICK_ENDDECLS
ELEMENT_REQUIRES(linuxmodule)
//Needed for Clicks compilation process.
EXPORT_ELEMENT(MyFirstElement)

