// ====================== /elements/local myFirstElement.hh ===============================
// 
// Written by: Alexander Jacobsen
// Element Description: Element which does nothing...
// Last Updated: 07.03.2011
//
// ========================================================================================

#ifndef CLICK_MYFIRSTELEMENT_HH
#define CLICK_MYFIRSTELEMENT_HH
#include <click/element.hh>
#include <click/straccum.hh>
#include <click/standard/storage.hh>


//All click declarations are enclosed within a macro pair, CLICK_DECLS and CLICKENDDECLS.
CLICK_DECLS
	
	// The class is defined in the header file (C++)
	class MyFirstElement : public Element, public Storage { public:
	//class MyFirstElement : public Element { public:
			MyFirstElement();
			~MyFirstElement();
			
			//Returns class name.
			const char *class_name() const { return "MyFirstElement"; }
			
			//Sets the elements ports. This element has one input and one output port.
			const char *port_count() const { return "1/1"; }

			//Sets the elements port behavior. This element has push ports.
			const char *processing() const { return  "h/lh"; }

			//configure() not defined. Default used.

			//initialize() not defined. Default used.
			int initialize(ErrorHandler*);

			void push(int port, Packet *p);
			Packet * pull(int);

			//Add a handler which makes it possible to read the nr of packets.
			void add_handlers();

			//inline Packet *	MyFirstElement::getFirstPacketToSend();

			int averageTimeBetweenAcks() const { return _averageTimeBetweenAcks; }

			//Function to put the packets into the queue.
			inline bool enqueue(Packet*);

			//Function which sets the leave time for each packet.
			inline void	setLeaveTimeForPacket(int);

			//Function which calculates the leave time for each packet.
			inline Timestamp calculateLeaveTimeForPacket();

			static String read_handler(Element*, void*);

			inline Packet* deq();

			void cleanup(CleanupStage);

			int calculateAverageBetweenACKs();

			void MyFirstElement::storeDestIPsToManipulate();

			inline bool MyFirstElement::checkIfDestAddressShouldBeManipulated();

			int MyFirstElement::partition(int start, int end);

			void MyFirstElement::swap(Timestamp &value1, Timestamp &value2, Packet *pack1, Packet *pack2);

			void MyFirstElement::quickSort(int start, int end);

			private:
	
			int _totalPacketCounter;
			int _tempPacketCounter;
			short _new_capacity;
			Packet* volatile * _q;
			StringAccum sa;
			
			//Dynamically allocate memory for the packet leave times.
			Timestamp *packetLeaveTime;
			String *listOfDestinationIPsToManipulate;
			String *kpMetrics;
			
			//Timestamps.
			Timestamp pullTimeStamp;
			Timestamp leaveTime;
			Timestamp timeCheck;
			Timestamp newTimeStamp;
			Timestamp oldTimeStamp;
			Timestamp totalTimeStampBetweenAcks;
			Timestamp tempTimeBetweenThisAndPreviousPack;

			//Offsets
			int _networkSourceAddrOffset;
			int _networkDestAddrOffset;
			uint8_t _annotationOffset;

			int _averageTimeBetweenAcks;

			int _drops;
			int _highwater_length;
			int _leaveTime;

			bool _increaseAckDelay;

			String _destinationIPs;
			int _nrOfIps;

			String _packDestinationAddress;



};

//Method which returns a packet on request (pull).
inline Packet *	MyFirstElement::deq()	{
	    int h = _head, t = _tail;

	    //Check the time now against the packet leave time to see if
	    //the packet should be released.
	    Timestamp now;
	    now.assign_now();

	    if (h != t && now>=packetLeaveTime[h])
	    {
	    	Packet *p = _q[h];
			packet_memory_barrier(_q[h], _head);
			_head = next_i(h);
			assert(p);

			sa << "------------- Packet nr: " << p->anno_s32(16)  <<" -------------------" << "\n";
			sa << "Time now : " << now.unparse() << "\n";
			sa << "Time to leave : " << packetLeaveTime[h].unparse() << "\n";

			click_chatter(sa.c_str());
			sa.clear();

			return p;

	    }
	    else
	    {
	    	//sa << "------------- Received Pull, no packs ready -------------------" << "\n";
	    	//click_chatter(sa.c_str());
	    	//sa.clear();
	    	return 0;
	    }
}

CLICK_ENDDECLS
#endif


