// ====================== elements/local/PacketDelayElement.cc ================
// 
// Written by: Alexander Jacobsen
// Element Description: Element which does nothing...
// Last Updated: 07.03.2011
//
// ========================================================================

// Mandatory inclusions. 
#include<click/config.h>
#include<click/confparse.hh>
#include<click/error.hh>
#include<click/string.hh>
#include<click/timestamp.hh>
#include <click/straccum.hh>
#include"PacketDelayElement.hh"






//All click declarations are enclosed within a macro pair, CLICK_DECLS and CLICKENDDECLS.
CLICK_DECLS
	
	PacketDelayElement::PacketDelayElement() : _primaryQueue(0), _primaryStorage(), _secondaryStorage(), _loggTimer(this)
	//PacketDelayElement::PacketDelayElement()
	{ 
		_totalPacketCounter = 0;
		_averagePacketCounter = 0;

		oldTimeStamp.assign_now();

		_primaryStorage.set_capacity(1000);
		_secondaryStorage.set_capacity(1000);
		//_head = 0;
		//_tail = 0;

		_drops = 0;
		//Set the leave time interval in ms.

		_networkDestAddrOffset = 16;
		
		_leaveTimeIncrease = 0;

		_loggInterval = 20;
		_leaveTime = 0;

		//Nr of IPs that id a flow.
		_nrOfIps = 0;

		//Sets the ip destination as id for a flow. Used for id and manipulation of
		//flows.
		_destinationIPs = "158.38.122.8:50,158.38.122.9:50,158.38.122.10:50,192.145.123.3:60,122.4.2.2:100,";

		_annotationOffset = 16;

		//Offset which returns the source ip address
		_networkSourceAddrOffset = 12;

		packetLeaveTime = new Timestamp[_primaryStorage.capacity()];

		listOfDestinationIPsToManipulate = new String[1000];

		listOfSourceIPsToManipulate = new String[1000];

		kpMetrics = new int[1000];

		_logg = "";

		_increaseAckDelay = true;

	}
	PacketDelayElement::~PacketDelayElement()
	{ 
	
		
	}

	static String
	averagecounter_read_time_between_ACKs_handler(Element *e)
	{
		PacketDelayElement *c = (PacketDelayElement *)e;
	  	return String(c->averageTimeBetweenAcks());

	}

	int PacketDelayElement::initialize(ErrorHandler *errh)
	{

		assert(!_primaryQueue && _primaryStorage.head() == 0 && _primaryStorage.tail() == 0);

		assert(!_secondaryQueue && _secondaryStorage.head() == 0 && _secondaryStorage.tail() == 0);
		//Array which stores pointers to each individual packet.
		_primaryQueue = (Packet **) CLICK_LALLOC(sizeof(Packet *) * (_primaryStorage.capacity() + 1));
		_secondaryQueue = (Packet **) CLICK_LALLOC(sizeof(Packet *) * (_secondaryStorage.capacity() + 1));
		
		//Initialize timer, mandatory.
		_loggTimer.initialize(this);

		//Fire the timer as soon as the router runs.
		_loggTimer.schedule_now();

		if (_primaryQueue == 0)
			return errh->error("out of memory");

		return 0;
	}

	//Add a handler which makes it possible to read the nr of packets.
	void PacketDelayElement::add_handlers() {
    	add_data_handlers("nrofpackets", Handler::OP_READ | Handler::OP_WRITE, &_totalPacketCounter);
    	add_data_handlers("logging", Handler::OP_READ | Handler::OP_WRITE, &_logg);
    	add_data_handlers("loggInterval", Handler::OP_READ | Handler::OP_WRITE, &_loggInterval);
    	add_data_handlers("packetDelay", Handler::OP_READ | Handler::OP_WRITE, &_leaveTime);
    	add_data_handlers("packetDelayIncrease", Handler::OP_READ | Handler::OP_WRITE, &_leaveTimeIncrease);
    	add_data_handlers("destinationIPs", Handler::OP_READ | Handler::OP_WRITE, &_destinationIPs);
		add_read_handler("averageTimeBetweenAcks", averagecounter_read_time_between_ACKs_handler, &_averageTimeBetweenAcks);
		//add_read_handler("length", read_handler, (void *)0);
	}

	//Defines the push method. This is where a packet arrives if an upstream element transfers the packet to
	//this element.
	void PacketDelayElement::push(int port, Packet *p) {

		//Sets the new timestamp to current time since 1970.
		newTimeStamp.assign_now();

		//Counts the number of packets traversing the element. One for total amount
		//of packets, the other used to calculate average over the last 1000 packets.
		_totalPacketCounter++;
		_averagePacketCounter++;

		//Adds the time between this and the previous packet to the
		//total time between packets.
		totalTimeStampBetweenAcks += (newTimeStamp-oldTimeStamp);

		//Timestamp used to print the time between this and previous packet.
		tempTimeBetweenThisAndPreviousPack = (newTimeStamp-oldTimeStamp);

		//Calculates the average time between acks.
		_averageTimeBetweenAcks = calculateAverageBetweenACKs();

		//Reset average time between acks calc every 1000 pack.
		if( (_averagePacketCounter%1000) ==0)
		{
			_averagePacketCounter = 1;
			totalTimeStampBetweenAcks.assign(0);
		}

		//Store the IPs in a array.
		storeDestIPsToManipulate();

		_packDestinationAddress = IPAddress(p->data() + _networkSourceAddrOffset).unparse();

		if( checkIfPacketBelongsToFlowOfInterest() ){

			sa << "************ Packet nr: " << String(_totalPacketCounter) << " RECOGNIZED ************" << "\n";

			//Prints the packet arrival time.
			sa << "Packet arrival time : " << newTimeStamp.sec() << " (sec) : " << newTimeStamp.msec()  << " (ms)\n";

			//Prints the difference between the arrival time of this and previous packet.
			sa << "Time between this and previous packet : " << tempTimeBetweenThisAndPreviousPack.unparse() << " (sec)\n";

			sa << "Total time between ACKs : " << totalTimeStampBetweenAcks.msec1() << " (ms)\n";

			//Prints the new timestamp in ms since 1970.
			sa << "Average Time between the last 1000 ACKs : ";
			sa << calculateAverageBetweenACKs() << " (ms)\n";

			sa << "Source IP Address = ";
			sa << IPAddress(p->data() + _networkDestAddrOffset) << "\n";

			sa << "Destination IP Address = ";
			sa << IPAddress(p->data() + _networkSourceAddrOffset) << "\n";

			//Prints amount of dropped packets from queue.
			sa << "Number of packets dropped: " << _drops << "\n";

			//Print the destination addresses of packets that should be manipulated.
			sa << "IP dest addr of packs to manipulate flow rate : " << _destinationIPs << "\n";
		}
		else{
			sa << "************ Packet nr: " << String(_totalPacketCounter) << " NOT RECOGNIZED ************" << "\n";
		}

		//Assigns the newTimestamp to OldTimestamp for use as reference for next arriving packet.
		oldTimeStamp = newTimeStamp;

		//Enqueue packet.
		enqueue(p);

		//Write debug text to the click chatter.
		click_chatter(sa.c_str());
		sa.clear();
	}

	inline bool	PacketDelayElement::enqueue(Packet *p)
	{
		//Check if the p actually points to a packet in memory. Used to avoid nullpoiters.
		assert(p);

		volatile int h, t;

		//Set packet nr, used for debugging.
		p->set_anno_s32( _annotationOffset, _totalPacketCounter );

		//Store the packet leave time in array. If packet should be delayed, new leave time is
		//calculated. Else the packet should leave ASAP.

		if( checkIfPacketBelongsToFlowOfInterest() ){

			//Packet is recognized.
			h = _primaryStorage.head(), t = _primaryStorage.tail();

			//Check for the next free index.
			//Returns t+1 if the next index is unoccupied.
			//Returns 0 if t = capacity of the queue.
			int nextUnoccupiedIndex = _primaryStorage.next_i(t);

			//If next free position in the queue is not the head position, store
			//the packet in the queue. Else, if the next free position is the head
			//(start) of the queue, the queue is full and the packet is dropped.
			if (nextUnoccupiedIndex != h) {

				//Put the packet in the position of the tail in the queue.
				//First time tail is 0, else it is the next unoccupied position.
				_primaryQueue[t] = p;

				//Set the time when the packet should leave the element.
				setLeaveTimeForPacket(t);

		    	_primaryStorage.packet_memory_barrier( _primaryQueue[t], t );
		    	_primaryStorage.set_tail( nextUnoccupiedIndex );

		    	//Gets the current size of the queue.
				//int s = size(h, nextUnoccupiedIndex);

				//if (s > _highwater_length)
				//_highwater_length = s;

		    	return true;
			}
			//Primary queue is full, drop packet
			else {

				p->kill();
				_drops++;

				return false;
			}

		}

		//If packet is not recognized. Store it in the secondary queue.
		else{

			h = _secondaryStorage.head(), t = _secondaryStorage.tail();

			//Check for the next free index.
			//Returns t+1 if the next index is unoccupied.
			//Returns 0 if t = capacity of the queue.
			int nextUnoccupiedIndex = _secondaryStorage.next_i(t);

			//If next free position in the queue is not the head position, store
			//the packet in the queue. Else, if the next free position is the head
			//(start) of the queue, the queue is full and the packet is dropped.
			if (nextUnoccupiedIndex != h) {

				//Put the packet in the position of the tail in the queue.
				//First time tail is 0, else it is the next unoccupied position.
				_secondaryQueue[t] = p;

				//Set the time when the packet should leave the element.
				//setLeaveTimeForPacket(t);

				_secondaryStorage.packet_memory_barrier( _secondaryQueue[t], t );
				_secondaryStorage.set_tail( nextUnoccupiedIndex );

				//Gets the current size of the queue.
				//int s = size(h, nextUnoccupiedIndex);

				//if (s > _highwater_length)
				//_highwater_length = s;

				return true;
			}
			//Primary queue is full, drop packet
			else {

				p->kill();
				_drops++;

				return false;
			}
		}
	}

	//Function which sets the when the packet should leave the element.
	//The inline keyword is used for small functions to enhance performance.
	inline void	PacketDelayElement::setLeaveTimeForPacket(int packetQueueIndex){
		Timestamp now;
		now.assign_now();

		int tempKpMetric = 0;

		bool calculateNewLeaveTime = false;

		//sa << "Time now (sec) : " << _leaveTime << " (ms) : " << now.msec() << "\n";

		//Get the KPmetric index of the flow of interest.
		tempKpMetric = getKpMetric();

		//Store the packet leave time in array. If packet should be delayed, new leave time is
		//calculated. Else the packet should leave ASAP.
		packetLeaveTime[packetQueueIndex] = calculateLeaveTimeForPacket(tempKpMetric);

		sa << "Packet leave time :  " << packetLeaveTime[packetQueueIndex].sec() << " (sec) : " << packetLeaveTime[packetQueueIndex].msec() << " (ms)\n";
		sa << "Packet recognized" << "\n";
		sa << "Queue head : " << _primaryStorage.head() << "\n";
		sa << "Queue tail : " << _primaryStorage.tail() << "\n";
	}

	//Function which checks if the destination of the address is recognized. If it is, then a new
	//leave time should be calculated. If not, set leavetime to now.
	inline bool PacketDelayElement::checkIfPacketBelongsToFlowOfInterest(){
			bool flowRecognized = false;

			//Loop through all destination ips if they exist
			if(_nrOfIps!=0){

				for(int i = 0; i<_nrOfIps; i++)
				{
						if( listOfDestinationIPsToManipulate[i].compare(_packDestinationAddress) == 0 )
						{
							flowRecognized = true;
						}
				}
			}

			return flowRecognized;
	}

	//Function which returns the KP metric index. Used to calculate the packet leave time.
	inline int PacketDelayElement::getKpMetric(){
		int kpMetric = 0;

		//Loop through all destination ips if they exist
		if(_nrOfIps!=0){

			for(int i = 0; i<_nrOfIps; i++)
			{
				if( listOfDestinationIPsToManipulate[i].compare(_packDestinationAddress) == 0 )
				{
					kpMetric = i;
				}
			}
		}

		return kpMetric;
	}

	//Function which calculates when the packet should leave the element.
	inline Timestamp PacketDelayElement::calculateLeaveTimeForPacket(int kpMetricIndex){
		//Sets the current time stamp.
		leaveTime.assign_now();

		//Leave time which is not read from handlers. Generated from KP metric.
		int _newLeaveTime = 0;

		_newLeaveTime = kpMetrics[kpMetricIndex];

		//Packet leave time should be generated from KPmetric.
		leaveTime = leaveTime.make_msec(leaveTime.sec() , ( leaveTime.msec()+_newLeaveTime));

		//Calculates new time when the packet should leave from delay read in handlers.
		//leaveTime = leaveTime.make_msec(leaveTime.sec() , ( leaveTime.msec()+_leaveTime));

		//if(_increaseAckDelay && _totalPacketCounter%100==0)
			//_leaveTime+=1;
		//else if(_leaveTime==150){
			//_increaseAckDelay = false;
			//_leaveTime = 0;
		//}

		return leaveTime;
	}

	//Function which clears the queues from memory.
	void PacketDelayElement::cleanup(CleanupStage)
	{
	    for (int i = _primaryStorage.head(); i != _primaryStorage.tail(); i = _primaryStorage.next_i(i)){
	    	_primaryQueue[i]->kill();

	    }

	    for (int i = _secondaryStorage.head(); i != _secondaryStorage.tail(); i = _secondaryStorage.next_i(i)){
	    	_secondaryQueue[i]->kill();

	   	}

	    CLICK_LFREE(_primaryQueue, sizeof(Packet *) * (_primaryStorage.capacity() + 1));
	    _primaryQueue = 0;

	    CLICK_LFREE(_secondaryQueue, sizeof(Packet *) * (_secondaryStorage.capacity() + 1));
	    _secondaryQueue = 0;

	    //Delete the array containing packets leave times.
	    delete [] packetLeaveTime;

	    //Make packetLeaveTime point to null
	    packetLeaveTime = 0;

	}

	//Method which is triggered when the element receives a request for a packet.
	Packet * PacketDelayElement::pull(int)
	{
	    //Fetch packet if there are any ready in primary queue. Else fetch from
		//secondary queue.
		bool _packetsInPrimaryQueue = false;
		bool _packetsInSecondaryQueue = false;
		Timestamp now;

		_packetsInPrimaryQueue = _checkForPacketsInQueue(&_primaryStorage);
		_packetsInSecondaryQueue = _checkForPacketsInQueue(&_secondaryStorage);

		//If there are packets in primary queue.
		if(_packetsInPrimaryQueue){

			now.assign_now();

			//Check if packet timestamp is reached. Else fetch packet from secondary queue.
			if (now>=packetLeaveTime[ _primaryStorage.head() ] ){
			//if(true){

				//Debug text.
				sa << "--------------------------------------------------------------------------" << "\n";
				sa << "-------------- Packet being fetched from PrimaryQueue --------------------" << "\n";
				sa << "Time to leave : " << packetLeaveTime[_primaryStorage.head()].unparse() << "\n";

				return decreaseQueue(&_primaryStorage, _primaryQueue);

			}
			else if(_packetsInSecondaryQueue){

				//Debug text.
				sa << "--------------------------------------------------------------------------" << "\n";
				sa << "------------- Packet being fetched from SecondaryQueue -------------------" << "\n";


				return decreaseQueue(&_secondaryStorage, _secondaryQueue);
			}
			else{
				//sa << "------------- Received Pull, no packs ready -------------------" << "\n";
				//click_chatter(sa.c_str());
				//sa.clear();
				return 0;
			}
		}

		//If there are no packets in primary queue. Fetch packet from secondary queue.
		else if (_packetsInSecondaryQueue) {

			//Debug text.
			sa << "--------------------------------------------------------------------------" << "\n";
			sa << "------------- Packet being fetched from SecondaryQueue -------------------" << "\n";


			return decreaseQueue(&_secondaryStorage, _secondaryQueue);
		}

		else{
			//sa << "------------- Received Pull, no packs ready -------------------" << "\n";
			//click_chatter(sa.c_str());
			//sa.clear();
			return 0;
		}
	}

	//Checks if head is equal to tail. If they are equal, then the
	//queue is empty.
	bool PacketDelayElement::_checkForPacketsInQueue(Storage* thisQueueControl){

		int _queueHead = thisQueueControl->head();
		int _queueTail = thisQueueControl->tail();

		if(_queueHead != _queueTail){
			return true;
		}
		else{
			return false;
		}
	}

	String PacketDelayElement::read_handler(Element *e, void *thunk)
	{
	    //PacketDelayElement *q = static_cast<PacketDelayElement *>(e);
	    //int which = reinterpret_cast<intptr_t>(thunk);
	    //switch (which) {
	    //  case 0:
		//return String(q->size());
	    //  default:
		return "";
	    //}
	}
	//Calculates the average time between incoming ACKs
	int PacketDelayElement::calculateAverageBetweenACKs(){

		//Adds the seconds from timestamp.
		int tempAvTimeBetweenAcksInMilliSecs = totalTimeStampBetweenAcks.msec1()/_averagePacketCounter;

		return tempAvTimeBetweenAcksInMilliSecs;
	}

	void PacketDelayElement::storeDestIPsToManipulate(){

		int stringSize = _destinationIPs.length();
		int separatorPos = 0;
		int separatorPosIterator = 0;
		int colonSeparatorPos = 0;

		String rest = _destinationIPs;
		String tempIP = "";
		String tempKpMetric = "";
		int tempMetric;

		_nrOfIps = 0;

		//Do as long as comma is found in string.
		while(separatorPos >= 0) {

			//Search for char in string. Returns -1 if the character is not found.
			colonSeparatorPos = rest.find_left(':' , 0);
			separatorPos = rest.find_left(',' , 0);

			//Do something if entries are found.
			if(separatorPos >= 0){

				//Update list of IPs
				tempIP = rest.substring(0,colonSeparatorPos);

				tempKpMetric = rest.substring(colonSeparatorPos+1, ( separatorPos-(colonSeparatorPos+1) ) );

				listOfDestinationIPsToManipulate[_nrOfIps] = tempIP;

				cp_integer(tempKpMetric, 0, &tempMetric);

				kpMetrics[_nrOfIps] = tempMetric;

				separatorPosIterator += separatorPos+1;

				rest = _destinationIPs.substring(separatorPosIterator,stringSize);

				stringSize = rest.length();

				_nrOfIps++;
			}
		}

		if( checkIfPacketBelongsToFlowOfInterest() ){

			//Print the found entries for debugging.
			for(int i = 0; i<_nrOfIps; i++){

				sa << "IP : " << listOfDestinationIPsToManipulate[i] << " KP metric : " <<  kpMetrics[i] << "\n";

			}
		}
	}

	//Sets the action of the timer.
	void PacketDelayElement::run_timer(Timer *timer){
		Timestamp timeNow;
		timeNow.assign_now();

		_logg += timeNow.unparse() + " : " + String(_leaveTime) + "\n";

		_leaveTime += _leaveTimeIncrease;

		_loggTimer.reschedule_after_sec(_loggInterval);
	}

CLICK_ENDDECLS
ELEMENT_REQUIRES(linuxmodule)
//Needed for Clicks compilation process.
EXPORT_ELEMENT(PacketDelayElement)

