// ====================== /elements/local PacketDelayElement.hh ===============================
// 
// Written by: Alexander Jacobsen
// Element Description: Element which does nothing...
// Last Updated: 07.03.2011
//
// ========================================================================================

#ifndef CLICK_PacketDelayElement_HH
#define CLICK_PacketDelayElement_HH
#include <click/element.hh>
#include <click/straccum.hh>
#include <click/standard/storage.hh>
#include <click/timer.hh>

//All click declarations are enclosed within a macro pair, CLICK_DECLS and CLICKENDDECLS.
CLICK_DECLS
	
	// The class is defined in the header file (C++)
	class PacketDelayElement : public Element { public:
	//class PacketDelayElement : public Element { public:
			PacketDelayElement();
			~PacketDelayElement();
			
			//Returns class name.
			const char *class_name() const { return "PacketDelayElement"; }
			
			//Sets the elements ports. This element has one input and one output port.
			const char *port_count() const { return "1/1"; }

			//Sets the elements port behavior. This element has push ports.
			const char *processing() const { return  "h/lh"; }

			//configure() not defined. Default used.

			//initialize() not defined. Default used.
			int initialize(ErrorHandler*);

			void push(int port, Packet *p);
			Packet * pull(int);

			//Add a handler which makes it possible to read the nr of packets.
			void add_handlers();

			int averageTimeBetweenAcks() const { return _averageTimeBetweenAcks; }

			//Function to put the packets into the queue.
			inline bool enqueue(Packet*);

			//Function which sets the leave time for each packet.
			inline void	setLeaveTimeForPacket(int);

			//Function which calculates the leave time for each packet.
			inline Timestamp calculateLeaveTimeForPacket(int);

			static String read_handler(Element*, void*);

			void cleanup(CleanupStage);

			int calculateAverageBetweenACKs();

			void PacketDelayElement::storeDestIPsToManipulate();

			inline bool PacketDelayElement::checkIfPacketBelongsToFlowOfInterest();

			void PacketDelayElement::run_timer(Timer *timer);

			inline Packet* decreaseQueue(Storage*, Packet** );

			bool _checkForPacketsInQueue(Storage*);

			inline int PacketDelayElement::getKpMetric();


			private:

			int _totalPacketCounter;
			int _averagePacketCounter;
			short _new_capacity;

			StringAccum sa;
			
			//Dynamically allocate memory for the packet leave times.
			Timestamp *packetLeaveTime;
			String *listOfDestinationIPsToManipulate;
			String *listOfSourceIPsToManipulate;
			int *kpMetrics;
			
			//Timestamps.
			Timestamp pullTimeStamp;
			Timestamp leaveTime;
			Timestamp timeCheck;
			Timestamp newTimeStamp;
			Timestamp oldTimeStamp;
			Timestamp totalTimeStampBetweenAcks;
			Timestamp tempTimeBetweenThisAndPreviousPack;

			//Offsets
			int _networkSourceAddrOffset;
			int _networkDestAddrOffset;
			uint8_t _annotationOffset;

			int _averageTimeBetweenAcks;

			int _drops;
			int _highwater_length;
			int _leaveTime;

			bool _increaseAckDelay;

			String _destinationIPs;
			int _nrOfIps;

			String _packDestinationAddress;

			String _logg;
			int _loggInterval;
			Timer _loggTimer;

			int _leaveTimeIncrease;

			//Storage controllers used to set the capacity, control where packets are
			//stored in the queue. Need one for each queue.
			Storage _primaryStorage;
			Storage _secondaryStorage;

			Packet* volatile * _primaryQueue;
			Packet* volatile * _secondaryQueue;

};

//Method which returns a packet on request (pull).
inline Packet *	PacketDelayElement::decreaseQueue(Storage* queueStorage, Packet** queue )	{

	    volatile int h = queueStorage->head(), t = queueStorage->tail();
	    Timestamp now;

	    now.assign_now();

	    //Check the time now against the packet leave time to see if
	    //the packet should be released.
	    if(h != t){

			Packet *p = queue[h];
			queueStorage->packet_memory_barrier(queue[h], h);
			//_primaryStorage.set_head( _primaryStorage.next_i(h) );

			queueStorage->set_head( queueStorage->next_i(h));

			assert(p);

			sa << "Packet nr: " << p->anno_s32(16)  << "\n";
			sa << "Time now : " << now.unparse() << "\n";
			sa << "--------------------------------------------------------------------------" << "\n";

			click_chatter(sa.c_str());
			sa.clear();

			return p;
	    }else
	    	return 0;
}

CLICK_ENDDECLS
#endif


