// ====================== elements/local/PacketDelayElement.cc ================
//
// Written by: Alexander Jacobsen
// Element Description: Element which rewrites the window field in the
//						TCP header. This element assume that window scaling
//						is not in use.
// Last Updated: 07.03.2011
//
// ========================================================================

// Mandatory inclusions.
#include<click/config.h>
#include <clicknet/tcp.h>
#include <clicknet/ip.h>
#include "TcpWindowFieldReWriter.hh"
#include <click/ipaddress.hh>
#include <click/confparse.hh>



#include<click/error.hh>
#include<click/string.hh>
#include<click/timestamp.hh>
#include <click/straccum.hh>




//All click declarations are enclosed within a macro pair, CLICK_DECLS and CLICKENDDECLS.
CLICK_DECLS

	TcpWindowFieldReWriter::TcpWindowFieldReWriter()
	{
		_packetCounter = 0;
		_newTcpWindowSize = -1;
		//_newTcpWindowSize = 0;

		listOfDestinationIPsToManipulate = new String[1000];
		listOfSourceIPsToManipulate = new String[1000];
		listOfRTTs = new int[1000];
		listOfTargetBitRates = new int[1000];

		//Nr of IPs that id a flow.
		nrOfServiceSessions = 0;

		serviceSessions = "129.241.197.92:158.38.14.136:0.98:9,129.241.197.91:158.38.14.135:0.56:99999,129.241.197.82:158.38.14.135:40:99999,129.241.197.82:158.38.14.136:40:99999,";



	}
	TcpWindowFieldReWriter::~TcpWindowFieldReWriter()
	{


	}

	int TcpWindowFieldReWriter::initialize(ErrorHandler *errh)
	{
		return 0;

	}

	//Add a handler which makes it possible to read the nr of packets.
	void TcpWindowFieldReWriter::add_handlers() {
		add_data_handlers("newTcpWindowSize", Handler::OP_READ | Handler::OP_WRITE, &_newTcpWindowSize);
		add_data_handlers("serviceSessions", Handler::OP_READ | Handler::OP_WRITE, &serviceSessions);
	}

	//Defines the push method. This is where a packet arrives if an element transfers the packet to
	//this element.
	Packet * TcpWindowFieldReWriter::simple_action(Packet *p){

		bool packBelongsToAServiceFlow = false;
		int newWindowSize = 0;
		int indexOfServiceFlow = 0;

		//Counter used for debug text.
		_packetCounter++;

		//Store all parameters retrieved from the service session handler.
		storeDestIPsToManipulateAndRTT();

		// TCP header
		click_tcp *tcpHeader = p->tcp_header();

		// IP header
		click_ip *ipHeader = p->ip_header();

		_currentTcpWindowSize = ntohs(tcpHeader->th_win);

		_ipSrcAddress = IPAddress(ipHeader->ip_src);
		_ipDestAddress = IPAddress(ipHeader->ip_dst);

		_debugText << "******************** Packet nr: " << _packetCounter << " **************************" << "\n";
		_ipHeader << "IP src adr: " << _ipSrcAddress << " --- IP dest adr: " << _ipDestAddress << "\n";

		_tcpNetworkHeader << "Old TCP Window size : " << _currentTcpWindowSize << "\n";

		//Update TCP window size if it is set in handler.
		if(_newTcpWindowSize>=0)
		{
			tcpHeader->th_win = ntohs(_newTcpWindowSize);
		}

		packBelongsToAServiceFlow = checkIfPacketBelongsToServiceFlow(_ipSrcAddress, _ipDestAddress);

		//Check if packet belongs to a service flow of interest. True: get the index of the parameters (ip, rtt, bitrate, etc) of this
		//service flow stored in arrays.
		if(packBelongsToAServiceFlow)
		{
			indexOfServiceFlow = getIndexOfServiceFlowInInformationArrays(_ipSrcAddress,_ipDestAddress);

			_debugText << "Source IP : " << listOfSourceIPsToManipulate[indexOfServiceFlow] << " Dest IP : " <<  listOfDestinationIPsToManipulate[indexOfServiceFlow] << " RTT : " << listOfRTTs[indexOfServiceFlow] <<" Target Rate : " << listOfTargetBitRates[indexOfServiceFlow] << "\n";

			newWindowSize = calculateWindowSize(listOfTargetBitRates[indexOfServiceFlow], listOfRTTs[indexOfServiceFlow] );

			_debugText << "New window size = " << newWindowSize << "\n";

			tcpHeader->th_win = ntohs(newWindowSize);


		}
		//else
		//{
				//_currentTcpWindowSize = ntohs(tcpHeader->th_win);
		//}


		unsigned plen = ntohs(ipHeader->ip_len) - (ipHeader->ip_hl << 2);
		unsigned packetCheckSum;

		tcpHeader->th_sum = 0;
		packetCheckSum = click_in_cksum((unsigned char *)tcpHeader, plen);
		tcpHeader->th_sum = click_in_cksum_pseudohdr(packetCheckSum, ipHeader, plen);

		_tcpNetworkHeader << "New TCP Window size : " << _currentTcpWindowSize << "\n";

		_tcpNetworkHeader << "ACK sequence nr : " << ntohs(tcpHeader->th_ack) << "\n";

		_tcpNetworkHeader << "ACK nr : " << (tcpHeader->th_ack) << "\n";


		//Write IP header to the click chatter.
		click_chatter(_ipHeader.c_str());
		_ipHeader.clear();

		//Write TCP network header to the click chatter.
		click_chatter(_tcpNetworkHeader.c_str());
		_tcpNetworkHeader.clear();

		//Write general debug text to the click chatter.
		click_chatter(_debugText.c_str());
		_debugText.clear();

		return p;

	}

	//Function which clears the queues from memory.
	void TcpWindowFieldReWriter::cleanup(CleanupStage)
	{
		//Delete arrays.
		//delete [] listOfDestinationIPsToManipulate;
		//delete [] listOfRTTs;
		//delete [] listOfSourceIPsToManipulate;
		//delete [] listOfTargetBitRates;
	}
	/**
	 * Function which retrieves all parameters from the service session
	 * handler and stores them in corresponding arrays.
	 */
	void TcpWindowFieldReWriter::storeDestIPsToManipulateAndRTT(){

			int _stringSize = serviceSessions.length();

			int _commaSeparatorPosIterator = 0;
			int _colonSeparatorPos = 0;
			int _startIndexOfDestIP = 0;
			int _lengthOfDestIP = 0;
			int _lengthOfRTT = 0;
			int _startIndexOfRTT = 0;
			int _previousColonSeparatorPos = 0;
			int _startIndexOfTargetBitRate = 0;
			int _lengthOfTargetBitRate = 0;


			String _rest = serviceSessions;
			String _sourceIP = "";
			String _destIP = "";
			String _tempKpTargetRate = "";
			String _tempKpRtt = "";

			long targetRate = 0;
			int _commaSeparatorPos = _rest.find_left(',' , 0);

			//Double parser not working as it should....
			//double rtt;

			int rtt;

			nrOfServiceSessions = 0;

			//Iterating the string of service session parameters. The string is on the following form:
			//sourceIP:destIP:RTT:targetBitRate,sourceIP:destIP:RTT:targetBitRate -->
			//The iteration runs as long as a comma is found in the string.
			while(_commaSeparatorPos > 0) {

				//Search for char in string. Returns -1 if the character is not found.
				_colonSeparatorPos = _rest.find_left(':' , 0);
				_commaSeparatorPos = _rest.find_left(',' , 0);

				//Do something if the handler contain service sessions.
				if(_commaSeparatorPos >= 0){

					//Extract source ip from handler string.
					_sourceIP = _rest.substring(0,_colonSeparatorPos);

					_previousColonSeparatorPos = _colonSeparatorPos;

					//Update colon separator with the position of the next colon separator.
					//(Here: after the destination address.)
					_colonSeparatorPos = _rest.find_left(':' , _previousColonSeparatorPos+1);

					_startIndexOfDestIP = _previousColonSeparatorPos+1;
					_lengthOfDestIP = _colonSeparatorPos-_startIndexOfDestIP;

					//Extract destination ip from handler string.
					_destIP = _rest.substring(_startIndexOfDestIP, _lengthOfDestIP);

					_previousColonSeparatorPos = _colonSeparatorPos;

					_colonSeparatorPos = _rest.find_left(':' , _previousColonSeparatorPos+1);

					//Update colon separator with the position of the next colon separator.
					//(Here: after the RTT.)
					_startIndexOfRTT = _previousColonSeparatorPos+1;
					_lengthOfRTT = _colonSeparatorPos-_startIndexOfRTT;

					_tempKpRtt = _rest.substring(_startIndexOfRTT, _lengthOfRTT );

					_previousColonSeparatorPos = _colonSeparatorPos;

					_startIndexOfTargetBitRate = _previousColonSeparatorPos+1;
					_lengthOfTargetBitRate = _commaSeparatorPos-_startIndexOfTargetBitRate;

					_tempKpTargetRate = _rest.substring(_startIndexOfTargetBitRate, _lengthOfTargetBitRate );


					//Store all parameters in corresponding arrays.
					listOfSourceIPsToManipulate[nrOfServiceSessions] = _sourceIP;
					listOfDestinationIPsToManipulate[nrOfServiceSessions] = _destIP;

					//Convert RTT from string to double.
					cp_integer(_tempKpRtt, 0, &rtt);
					listOfRTTs[nrOfServiceSessions] = rtt;

					//Convert target rate from string to int.
					cp_integer(_tempKpTargetRate, 0, &targetRate);
					listOfTargetBitRates[nrOfServiceSessions] = targetRate;

					_commaSeparatorPosIterator += _commaSeparatorPos+1;

					_rest = serviceSessions.substring(_commaSeparatorPosIterator,_stringSize);

					_stringSize = _rest.length();

					nrOfServiceSessions++;
				}
			}

		}
		//Function which checks if the destination of the address is recognized. If it is, then a new
		//leave time should be calculated. If not, set leavetime to now.
		inline bool TcpWindowFieldReWriter::checkIfPacketBelongsToFlowOfInterest() {
			bool _flowRecognized = false;

			//Loop through all destination ips if they exist
			if (nrOfServiceSessions != 0) {

				for (int i = 0; i < nrOfServiceSessions; i++) {
					if (listOfDestinationIPsToManipulate[i].compare(
							_ipDestAddress) == 0) {
						_flowRecognized = true;
					}
				}
			}

			return _flowRecognized;
		}

		//Function which checks if the destination of the address is recognized. If it is, then a new
		//leave time should be calculated. If not, set leavetime to now.
		inline bool TcpWindowFieldReWriter::checkIfPacketBelongsToServiceFlow(String sourceIP, String destIp) {
			bool flowRecognized = false;

			//Loop through all destination ips if they exist
			if (nrOfServiceSessions != 0) {

				for (int i = 0; i < nrOfServiceSessions; i++) {
					if ( listOfSourceIPsToManipulate[i].compare(sourceIP) == 0 && listOfDestinationIPsToManipulate[i].compare(destIp) == 0 ) {
						flowRecognized = true;
					}
				}
			}

			return flowRecognized;
		}

		inline int TcpWindowFieldReWriter::getIndexOfServiceFlowInInformationArrays(String sourceIP, String destIp)
		{
			int serviceFlowIndex = -1;

			//Loop through all destination ips if they exist
			if (nrOfServiceSessions != 0) {

				for (int i = 0; i < nrOfServiceSessions; i++)
				{
					if ( listOfSourceIPsToManipulate[i].compare(sourceIP) == 0 && listOfDestinationIPsToManipulate[i].compare(destIp) == 0)
					{
						serviceFlowIndex = i;
					}
				}
			}

			return serviceFlowIndex;
		}


		//Function which calculates new window size given bit rate and RTT
		inline long TcpWindowFieldReWriter::calculateWindowSize(long newBitRate, int RTT)
		{
			long newWindowSize = 0;
			_debugText << "Parameters in calculateWinSize : BitRate = " << newBitRate << " RTT = " << RTT << "\n";

			//Tweak
			//newBitRate = newBitRate/120;

			newWindowSize = (newBitRate*RTT)/(8*1000);
			_debugText << "Parameters in calculateWinSize : newWinSize = " << newWindowSize << "\n";
			return newWindowSize;
		}




CLICK_ENDDECLS
ELEMENT_REQUIRES(linuxmodule)
//Needed for Clicks compilation process.
EXPORT_ELEMENT(TcpWindowFieldReWriter)

