// ====================== /elements/local PacketDelayElement.hh ===============================
//
// Written by: Alexander Jacobsen
// Element Description: Element which does nothing...
// Last Updated: 13.04.2011
//
// ========================================================================================

#ifndef CLICK_TcpWindowFieldReWriter_HH
#define CLICK_TcpWindowFieldReWriter_HH

#include <click/element.hh>
#include <click/straccum.hh>



//All click declarations are enclosed within a macro pair, CLICK_DECLS and CLICKENDDECLS.
CLICK_DECLS

	// The class is defined in the header file (C++)
	class TcpWindowFieldReWriter : public Element { public:

			TcpWindowFieldReWriter();
			~TcpWindowFieldReWriter();

			//Returns class name.
			const char *class_name() const { return "PacketDelayElement"; }

			//Sets the elements ports. This element has one input and one output port.
			const char *port_count() const { return PORTS_1_1X2; }

			//Sets the elements port behavior. This element has push ports.
			const char *processing() const { return PROCESSING_A_AH; }

			//configure() not defined. Default used.

			//initialize() not defined. Default used.
			int initialize(ErrorHandler*);

			Packet * TcpWindowFieldReWriter::simple_action(Packet *p);

			//Add a handler which makes it possible to read the nr of packets.
			void add_handlers();
			//static String read_handler(Element*, void*);

			void TcpWindowFieldReWriter::cleanup(CleanupStage);

			void TcpWindowFieldReWriter::storeDestIPsToManipulate();

			inline bool TcpWindowFieldReWriter::checkIfPacketBelongsToFlowOfInterest();

			void TcpWindowFieldReWriter::storeDestIPsToManipulateAndRTT();

			inline long TcpWindowFieldReWriter::calculateWindowSize(long, int);

			inline bool TcpWindowFieldReWriter::checkIfPacketBelongsToServiceFlow(String sourceIP, String DestIp);

			inline int TcpWindowFieldReWriter::getIndexOfServiceFlowInInformationArrays(String sourceIP, String DestIp);

			private:

			StringAccum _debugText, _tcpNetworkHeader, _ipHeader;

			long int _packetCounter;

			unsigned _currentTcpWindowSize;

			//long int
			int _newTcpWindowSize;

			//Dynamically allocate memory for the packet leave times.
			String *listOfDestinationIPsToManipulate;
			String *listOfSourceIPsToManipulate;
			int *listOfRTTs;
			long *listOfTargetBitRates;

			int nrOfServiceSessions;

			String serviceSessions;

			String _ipSrcAddress, _ipDestAddress;

};

CLICK_ENDDECLS
#endif


