package no.ntnu.item.router;

import java.io.IOException;

import no.ntnu.item.router.utilities.ActionPlaneSessions;
import no.ntnu.item.router.utilities.GuiXmlWriter;
import no.ntnu.item.router.utilities.Parser;
import no.ntnu.item.router.utilities.WebServer;
import static no.ntnu.item.router.Constants.WEB_SERVER_ADDR;
import static no.ntnu.item.router.Constants.WEB_SERVER_PORT;

/**
 * The main class.
 * 
 * @author Kristian Haugene
 * 
 */
public class RouterControl {

	/** Controls the {@link ServiceSession} objects known by the router */
	private ServiceSessionBank serviceBank;

	/** A {@link GuiXmlWriter} object that will keep the GUI up to date */
	private GuiXmlWriter guiXmlWriter;

	/** The {@link ActionPlaneSessions} will control AP relevant sessions */
	private ActionPlaneSessions actionPlaneSessions;

	/**
	 * The {@link ActionReasoner} will continuously communicate with AP and set
	 * params
	 */
	private ActionReasoner actionReasoner;

	/** Holds a {@link Parser} object */
	private Parser parser;

	/**
	 * Main method to start the {@link RouterControl}
	 * 
	 * @param args
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		new RouterControl();
	}

	/**
	 * Constructor<br>
	 * Init objects
	 * 
	 * @throws IOException
	 */
	public RouterControl() throws IOException {

		parser = new Parser();
		guiXmlWriter = new GuiXmlWriter();
		actionReasoner = new ActionReasoner();
		actionPlaneSessions = new ActionPlaneSessions(actionReasoner);
		serviceBank = new ServiceSessionBank(parser, guiXmlWriter,
				actionPlaneSessions);
		new WebServer(serviceBank);

		System.out.println("Router Control Java Application is running");
		System.out.println("Web server running at " + WEB_SERVER_ADDR + ":"
				+ WEB_SERVER_PORT);
	}

}
