package no.ntnu.item.router;

/**
 * Information object that contains a subset of the information in a
 * {@link ServiceSession}. This subset is the values relevant to the Click
 * handler. A list of this object will contain all the information that is to be
 * written to Click so that Click can perform action on the stream.
 * 
 * @author Kristian Haugene
 * 
 */
public class SessionHandlerInfo {

	/** The IP address of the host running the Silverlight Client */
	private String clientAddress;

	/** The IP address of the media server providing content */
	private String sourceAddress;

	/** The latest ping value of the session */
	private Double ping;

	/** The rate determined that Click should try to obtain */
	private Integer clickTargetRate;

	/**
	 * Parameterized constructor
	 * 
	 * @param sourceAddress
	 * @param clientAddress
	 * @param ping
	 * @param clickTargetRate
	 */
	public SessionHandlerInfo(String clientAddress, String sourceAddress,
			Double ping, Integer clickTargetRate) {
		this.clientAddress = clientAddress;
		this.sourceAddress = sourceAddress;
		this.ping = ping;
		this.clickTargetRate = clickTargetRate;
	}

	/**
	 * Getter
	 * 
	 * @return the Silverlight Client host IP address
	 */
	public String getClientAddress() {
		return clientAddress;
	}

	/**
	 * Getter
	 * 
	 * @return the content delivery server IP address
	 */
	public String getSourceAddress() {
		return sourceAddress;
	}

	/**
	 * Getter
	 * 
	 * @return the ping measurement for the session
	 */
	public Double getPing() {
		return ping;
	}

	/**
	 * Getter
	 * 
	 * @return the rate Click should try to obtain
	 */
	public Integer getClickTargetRate() {
		return clickTargetRate;
	}
}
