package no.ntnu.item.router.utilities;

import java.util.ArrayList;
import java.util.List;

import static no.ntnu.item.router.Constants.SESSION_INACTIVE_LIMIT;

import no.ntnu.item.router.ServiceSession;
import no.ntnu.item.router.ServiceSessionBank;
import no.ntnu.item.router.SilverlightMeasurement;

/**
 * A separate thread that is responsible for removing {@link ServiceSession}
 * objects from the {@link ServiceSessionBank} if they are idle.
 * 
 * @author Kristian Haugene
 * 
 */
public class ServiceSessionBankCleaner extends Thread {

	/** Holds the {@link ServiceSessionBank} */
	private ServiceSessionBank bank;

	/**
	 * Constructor, starts the Thread
	 * 
	 * @param serviceSessionBank
	 *            the {@link ServiceSessionBank} to watch
	 */
	public ServiceSessionBankCleaner(ServiceSessionBank serviceSessionBank) {
		this.bank = serviceSessionBank;
		this.start();
	}

	public void run() {

		while (true) {

			// Clean every 30 seconds
			try {
				Thread.sleep(30 * 1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			List<String> sessionsToRemove = findInactiveSessions();
			removeInactiveSessions(sessionsToRemove);
			
			// We have changed list of sessions
			if(!sessionsToRemove.isEmpty()){
				bank.updateGuiXmlAndApSessions();
			}

		}

	}
	
	private void removeInactiveSessions(List<String> sessionsToRemove) {
		
		// Get the list from the ServiceSessionBank
		List<ServiceSession> sessionList = bank.getServiceSessions();
		
		// Iterate the sessions to remove
		for(String toRemove : sessionsToRemove){
			
			// For each, find it and remove it
			for(int i = 0; i< sessionList.size();i++){
				if(sessionList.get(i).getId().equals(toRemove)){
					System.out.println("Removed: " + sessionList.get(i).getId());
					sessionList.remove(i);
					break;
				}
			}
		}
		
	}

	private List<String> findInactiveSessions() {
		
		// Maintain a list of inactive sessions
		List<String> sessionIdsToRemove = new ArrayList<String>();
		
		// Get the list from the ServiceSessionBank
		List<ServiceSession> sessionList = bank.getServiceSessions();
		
		//Iterate the list, remember inactive sessions
		for(ServiceSession session : sessionList){
			
			if(haveNoRecentMeasurements(session)){
				sessionIdsToRemove.add(session.getId());
			}
		}
		
		// Return it
		return sessionIdsToRemove;
		
	}

	/** 
	 * Checks if the given session have been inactive for long enough to remove it.
	 * @param session
	 * @return true if inactive, false otherwise
	 */
	private boolean haveNoRecentMeasurements(ServiceSession session) {
		
		// Get the current time
		Long timeNow = System.currentTimeMillis();
		
		// Get the latest measurement
		List<SilverlightMeasurement> measurements = session.getMeasurements();
		SilverlightMeasurement lastMeasurement = measurements.get(measurements.size()-1);
		
		// Calculate time since last measurement
		Long timeSinceLastMeasurement = timeNow - lastMeasurement.getRouterTimeStamp();
		
		return timeSinceLastMeasurement > SESSION_INACTIVE_LIMIT;
		
	}

}
