package no.ntnu.item.test;

import static org.junit.Assert.assertEquals;
import no.ntnu.item.router.ActionReasoner;
import no.ntnu.item.router.ServiceSessionBank;
import no.ntnu.item.router.utilities.ActionPlaneSessions;
import no.ntnu.item.router.utilities.GuiXmlWriter;
import no.ntnu.item.router.utilities.Parser;

import org.junit.Test;

public class ServiceSessionBankTest {
	
	@Test
	public void TestReferencesAndInstantiation(){
		Parser parser = new Parser();
		GuiXmlWriter guiXmlWriter = new GuiXmlWriter();
		ActionReasoner actionReasoner = new ActionReasoner();
		ActionPlaneSessions actionPlaneSessions = new ActionPlaneSessions(actionReasoner);
		
		new ServiceSessionBank(parser, guiXmlWriter, actionPlaneSessions);
	}
	
	@Test
	public void FirstMeasurementShouldCreateNewSessionAndBeStored(){
		
		Parser parser = new Parser();
		GuiXmlWriter guiXmlWriter = new GuiXmlWriter();
		ActionReasoner actionReasoner = new ActionReasoner();
		ActionPlaneSessions actionPlaneSessions = new ActionPlaneSessions(actionReasoner);
		
		ServiceSessionBank ssb = new ServiceSessionBank(parser, guiXmlWriter, actionPlaneSessions);
		
		String measurement = "event=InitialInformation&eventMessage=DONE&playState=Playing&fullScreen=False&clientId=64ece20e-0d1d-48d2-aac3-6fba5d728d65&avgProcessorLoad=49,5&avgProcessLoad=28,8&videoDownloadBitrate=2436000&avaliableVideoBitrates=300000 - 427000 - 608000 - 866000 - 1233000 - 1636000 - 2436000&myIp=129.241.197.86&ispAS=AS224&sourceIpAndPort=video3.smoothhd.com.edgesuite.net:80&timeIncrement=1,95";
		
		// Register
		ssb.registerMeasurement(measurement);
		
		// Now check that it is registered
		assertEquals(1,ssb.getServiceSessions().size());
		
		// Check that it looks OK
		assertEquals("64ece20e-0d1d-48d2-aac3-6fba5d728d65", ssb.getServiceSessions().get(0).getId());
		assertEquals("DONE", ssb.getServiceSessions().get(0).getLatestMeasurement().getEventMessage());
	}
	
	@Test
	public void AddingTwoMeasurementsWithDifferentIdShouldCreateTwoSessions(){
		
		Parser parser = new Parser();
		GuiXmlWriter guiXmlWriter = new GuiXmlWriter();
		ActionReasoner actionReasoner = new ActionReasoner();
		ActionPlaneSessions actionPlaneSessions = new ActionPlaneSessions(actionReasoner);
		
		ServiceSessionBank ssb = new ServiceSessionBank(parser, guiXmlWriter, actionPlaneSessions);
		
		String measurement1 = "event=InitialInformation&eventMessage=DONE&playState=Playing&fullScreen=False&clientId=64ece20e-0000-48d2-aac3-6fba5d728d65&avgProcessorLoad=49,5&avgProcessLoad=28,8&videoDownloadBitrate=2436000&avaliableVideoBitrates=300000 - 427000 - 608000 - 866000 - 1233000 - 1636000 - 2436000&myIp=129.241.197.86&ispAS=AS224&sourceIpAndPort=video3.smoothhd.com.edgesuite.net:80&timeIncrement=1,95";
		String measurement2 = "event=InitialInformation&eventMessage=DONE&playState=Playing&fullScreen=False&clientId=64ece20e-1111-48d2-aac3-6fba5d728d65&avgProcessorLoad=49,5&avgProcessLoad=28,8&videoDownloadBitrate=2436000&avaliableVideoBitrates=300000 - 427000 - 608000 - 866000 - 1233000 - 1636000 - 2436000&myIp=129.241.197.86&ispAS=AS224&sourceIpAndPort=video3.smoothhd.com.edgesuite.net:80&timeIncrement=1,95";
		
		// Register
		ssb.registerMeasurement(measurement1);
		ssb.registerMeasurement(measurement2);
		
		// Now check that it is registered
		assertEquals(2,ssb.getServiceSessions().size());
		
		// Check that it looks OK
		assertEquals("64ece20e-0000-48d2-aac3-6fba5d728d65", ssb.getServiceSessions().get(0).getId());
		assertEquals("64ece20e-1111-48d2-aac3-6fba5d728d65", ssb.getServiceSessions().get(1).getId());
	}
	
	@Test
	public void WhenAUserProvidesATargetBitRateItShouldBeSetToTheCorrectServiceSession(){
		
		Parser parser = new Parser();
		GuiXmlWriter guiXmlWriter = new GuiXmlWriter();
		ActionReasoner actionReasoner = new ActionReasoner();
		ActionPlaneSessions actionPlaneSessions = new ActionPlaneSessions(actionReasoner);
		
		ServiceSessionBank ssb = new ServiceSessionBank(parser, guiXmlWriter, actionPlaneSessions);
		
		String measurement1 = "event=InitialInformation&eventMessage=DONE&playState=Playing&fullScreen=False&clientId=64ece20e-0000-48d2-aac3-6fba5d728d65&avgProcessorLoad=49,5&avgProcessLoad=28,8&videoDownloadBitrate=2436000&avaliableVideoBitrates=300000 - 427000 - 608000 - 866000 - 1233000 - 1636000 - 2436000&myIp=129.241.197.86&ispAS=AS224&sourceIpAndPort=video3.smoothhd.com.edgesuite.net:80&timeIncrement=1,95";
		String measurement2 = "event=InitialInformation&eventMessage=DONE&playState=Playing&fullScreen=False&clientId=64ece20e-1111-48d2-aac3-6fba5d728d65&avgProcessorLoad=49,5&avgProcessLoad=28,8&videoDownloadBitrate=2436000&avaliableVideoBitrates=300000 - 427000 - 608000 - 866000 - 1233000 - 1636000 - 2436000&myIp=129.241.197.86&ispAS=AS224&sourceIpAndPort=video3.smoothhd.com.edgesuite.net:80&timeIncrement=1,95";
		
		// Register
		ssb.registerMeasurement(measurement1);
		ssb.registerMeasurement(measurement2);
		
		/*
		 * We have two measurements, lets register a target bit rate for one of them 
		 */
		
		ssb.registerUserPreferences("64ece20e-0000-48d2-aac3-6fba5d728d65=6666");
		
		assertEquals(6666, ssb.getServiceSessions().get(0).getTargetVideoBitrate().intValue());
		assertEquals(0, ssb.getServiceSessions().get(1).getTargetVideoBitrate().intValue());
		
		/*
		 * Now change both of them, they should both have a target value 
		 */
		
		ssb.registerUserPreferences("64ece20e-0000-48d2-aac3-6fba5d728d65=1234&64ece20e-1111-48d2-aac3-6fba5d728d65=7777");

		assertEquals(1234, ssb.getServiceSessions().get(0).getTargetVideoBitrate().intValue());
		assertEquals(7777, ssb.getServiceSessions().get(1).getTargetVideoBitrate().intValue());
	}
}
