package no.ntnu.item.test.utilities;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.HashMap;

import no.ntnu.item.router.SilverlightMeasurement;
import no.ntnu.item.router.enums.PlayState;
import no.ntnu.item.router.utilities.Parser;

import org.junit.Test;

public class ParserTest {

	
	@Test
	public void TestReferences(){
		@SuppressWarnings("unused")
		Parser p = new Parser();
	}
	
	@Test
	public void ParserShouldReturnASilverlightMeasurementWhenGivenStringRepresentation() throws UnknownHostException{
		Parser p = new Parser();
		String measurement = "event=InitialInformation&eventMessage=DONE&playState=Playing&fullScreen=False&clientId=64ece20e-0d1d-48d2-aac3-6fba5d728d65&avgProcessorLoad=49,5&avgProcessLoad=28,8&videoDownloadBitrate=2436000&avaliableVideoBitrates=300000 - 427000 - 608000 - 866000 - 1233000 - 1636000 - 2436000&myIp=129.241.197.86&ispAS=AS224&sourceIpAndPort=video3.smoothhd.com.edgesuite.net:80&timeIncrement=1,95";
		
		SilverlightMeasurement sm = p.getSilverlightMeasurementFromString(measurement);
		
		/*
		 * Assert that the returned SilverlightMeasurement has all fields set correctly
		 */
		
		assertEquals("InitialInformation", sm.getEvent());
		assertEquals("DONE", sm.getEventMessage());
		assertEquals(PlayState.PLAYING, sm.getPlayState());
		assertEquals(Boolean.FALSE, sm.getFullScreen());
		assertEquals("64ece20e-0d1d-48d2-aac3-6fba5d728d65", sm.getClientId());
		assertTrue("Processor load is wrong",sm.getAvgProcessorLoad().equals(new Double(49.5)));
		assertTrue("Process load is wront",sm.getAvgProcessLoad().equals(new Double(28.8)));
		assertTrue("Avaliable bitrates are wrong length",sm.getVideoDownloadBitrate().equals(new Integer(2436000)));
		assertEquals(7, sm.getAvaliableVideoBitrates().size());
		assertEquals(300000, sm.getAvaliableVideoBitrates().get(0).intValue());
		assertEquals(2436000, sm.getAvaliableVideoBitrates().get(6).intValue());
		assertEquals("AS224", sm.getIspAS());
		assertTrue("The timeIncrement is wrong", sm.getTimeIncrement().equals(new Double(1.95)));
		
		InetAddress myIpAdr = InetAddress.getByName("129.241.197.86");
		InetSocketAddress adr = new InetSocketAddress("video3.smoothhd.com.edgesuite.net", 80);
		
		assertTrue("myIp does not match", myIpAdr.equals(sm.getMyIp()));
		assertTrue("sourceIpAndPort is wrong", adr.equals(sm.getSourceIpAndPort()));
	}
	
	@Test
	public void ParserShouldReturnAHashMapWithUserPreferencesWhenGivenCorrectInputString(){
		Parser p = new Parser();
		String userPreferences = "client-1=5215&hex-dec-rep=65498&myhex=6518";
		
		HashMap<String, Integer> theMap = p.getUserPreferencesMapFromString(userPreferences);
		
		// Check that the size is right
		assertEquals("Size of map is not correct!", 3, theMap.size());
		
		// Check that all keys exist in Map
		assertTrue(theMap.containsKey("client-1"));
		assertTrue(theMap.containsKey("hex-dec-rep"));
		assertTrue(theMap.containsKey("myhex"));
		
		// Check that all values are right
		assertEquals(5215, theMap.get("client-1").intValue());
		assertEquals(65498, theMap.get("hex-dec-rep").intValue());
		assertEquals(6518, theMap.get("myhex").intValue());
		
	}
}
