package no.ntnu.item.batchSqlInsert;

import java.io.BufferedReader;
import java.io.FileReader;
import java.sql.*;
import java.util.ArrayList;
import java.util.List;

/**
 * A simple class for inserting to statistics database
 * 
 * @author Kristian Haugene
 * 
 */
public class BatchControl {

	/** Holds the statement object used to execute queries */
	private Statement statement;

	/** Holds connection object that is the connection to the database */
	private Connection connection;
	
	/** The parser that makes Sql INSERT queries */
	private Parser parser;

	/**
	 * Main method
	 * @param args
	 */
	public static void main(String[] args) {
		
		// Check that we have a parameter
		if(args.length == 0 || args[0] == null || args[0].equals("")){
			System.out.println("No input parameter given");
			return;
		}
		
		// START
		new BatchControl(args[0]);


	}
	
	/**
	 * Constructor Reads a log file given by the date parameter<br>
	 * Sets up a connection to the database<br>
	 * Parses the log file to sql queries and executes them<br>
	 * Closes the connection<br>
	 * 
	 * @param date
	 *            the date to retrieve log from
	 */
	public BatchControl(String date){
		
		parser = new Parser();
		
		// Make filename from date
		String fileName = "log"+date+".txt";
		
		// Read log entries
		List<String> logEntries = readFile("/var/www/logger/logs/"+ fileName);

		// Connect to database
		this.setUpConnectionToDatabase();
		
		// For each logEntry, parse it and run the Sql INSERT
		for(String logEntry : logEntries){
			insertToDatabase(parser.getSqlInsertQueryFromLogEntry(logEntry, date));
		}

		// Close it
		this.closeStatementAndConnection();
		
		/*
		 * AND WE ARE DONE
		 */
		System.out.println("DONE");
	
	}

	private void insertToDatabase(String QueryString) {
		try {

			int updateQuery = 0;
			updateQuery = statement.executeUpdate(QueryString);

			// updateQuery should be 1, as 1 line was inserted. If 0, something is wrong
			if (updateQuery == 0) {
				System.err.println(QueryString + "COULD NOT BE INSERTED");
			}
			
		} catch (Exception e) {
			e.printStackTrace();
			System.out.println("Exception in insertToDatabase()");
		}
	}

	private void setUpConnectionToDatabase() {
		try {
			/*
			 * Create string of connection url within specified format with
			 * machine name, port number and database name. Here machine name id
			 * localhost and database name is usermaster.
			 */
			String connectionURL = "jdbc:mysql://localhost:3306/statistics";

			// declare a connection by using Connection interface
			connection = null;

			// declare object of Statement interface that uses for executing sql
			// statements.
			statement = null;

			// Load JBBC driver "com.mysql.jdbc.Driver".
			Class.forName("com.mysql.jdbc.Driver").newInstance();

			/*
			 * Create a connection by using getConnection() method that takes
			 * parameters of string type connection url, user name and password
			 * to connect to database.
			 */
			connection = DriverManager.getConnection(connectionURL,
					"batchProcess", "julekalender");

			/*
			 * createStatement() is used for create statement object that is
			 * used for sending sql statements to the specified database.
			 */
			statement = connection.createStatement();

		} catch (Exception e) {
			e.printStackTrace();
			System.out.println("Exception in setUpConnectionToDatabase()");
		}
	}

	private void closeStatementAndConnection() {
		try {
			// close all the connections.
			statement.close();
			connection.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Reads the file at the given path. Returns the file as a list of strings.
	 * Every line is an element in the list.
	 * 
	 * @param filename
	 *            path to file
	 * @return lines in file, in list
	 */
	private static List<String> readFile(String filename) {

		String line = "";
		List<String> linesInConfigFile = new ArrayList<String>();			
	  
		try {				
			BufferedReader reader = new BufferedReader(new FileReader(filename));
	        
			while ((line = reader.readLine()) != null){
	        	linesInConfigFile.add(line);
	        }
	        reader.close();
	       		   		        
		} catch (Exception e){
			e.printStackTrace();
			System.out.println("ERROR IN READING FILE");
		}
		return linesInConfigFile;
	
	}

}
