﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Threading;
using System.Collections.ObjectModel;

namespace ReportingSilverlightApp
{
    public partial class MainPage : UserControl
    {

        // Represents a globally unique identifier (GUID).
        Guid _guid;

        // System Analytics object to read CPU information
        private static System.Windows.Analytics _analytics;

        // Init placeholders for gui objects
        private static System.Windows.Controls.TextBlock _infoBlock;
        private static Microsoft.SilverlightMediaFramework.Core.SMFPlayer _myPlayer;

        // This field holds a boolean value for determining if the init process of the player is done(the trancient period)
        private bool _initDone = false;

        // Information holders
        private String _eventCause = "";
        private String _eventMessage = "";
        private String _avaliableVideoBitrates = "";
        private String _myIp = "";
        private String _ispAS = "";
        private String _videoBitrate = "";
        private String _previousVideoBitrate = "0";
        private String _latestPlayState = "";
        private String _fullScreen = "False";
        private DateTime _clientStartUpTime;

        // WebClient for doing http requests, and result strings for storing async responses
        private static WebClient _myClient;
        private String _httpDownResult = "";
        private String _httpUpResult = "";

        // AutoResetEvent to alert the sending thread that new events have been added to the queue
        static AutoResetEvent autoEvent;

        // A List functioning as a queue for events that have occured
        private List<PostInformation> eventQ;

        public MainPage()
        {
            InitializeComponent();

            // Create the GUID
            _guid = Guid.NewGuid();

            // Assign variables to UI elements

            Microsoft.Web.Media.SmoothStreaming.SmoothStreamingMediaElement hei = new Microsoft.Web.Media.SmoothStreaming.SmoothStreamingMediaElement();
            
            _myPlayer = mediaPlayer;
            _infoBlock = infoBlock;

            // Init webclient
            _myClient = new WebClient();

            // Remember the time the application was initialized
            _clientStartUpTime = DateTime.UtcNow;

            // Init analytics object
            _analytics = new System.Windows.Analytics();

            // For peace
            _myPlayer.IsMuted = true;

            // Display debug text for users that are outside the click test bed
            SetDebugTextBlock("Debug text should appear here if you're behind the click test bed. If not, you have to check the logs for clientId: " + _guid);

            // Init the eventQ
            eventQ = new List<PostInformation>();

            // Init thread sync
            autoEvent = new AutoResetEvent(false);

            // Retrieve initial information
            StartInitThread();

            // Start the eventProcessing thread
            StartInfoThread();

            // Start the thread for periodic "keep-alive" messages
            StartPeriodicMeasurmentThread();

        }

        /**
         * This method is called when the DownloadBitrateChanged event is triggered in the UI object
         */
        private void mediaPlayer_DownloadBitrateChanged(object sender, Microsoft.SilverlightMediaFramework.Core.CustomEventArgs<long> e)
        {
            _videoBitrate = e.Value.ToString();

            // Do not report before init time has passed
            if (!_initDone)
            {
                return;
            }

            // Register that the current event is a DownloadBitrateChanged
            RegisterEvent(false, true, false, false, false);

        }

        /**
         * This method is called when the PlayStateChanged event is triggered in the UI object
         */
        private void mediaPlayer_PlayStateChanged(object sender, Microsoft.SilverlightMediaFramework.Core.CustomEventArgs<Microsoft.SilverlightMediaFramework.Plugins.Primitives.MediaPluginState> e)
        {
            _latestPlayState = e.Value.ToString();

            // Do not report before init time has passed
            if (!_initDone)
            {
                return;
            }

            // Register that the current event is a PlayStateChanged
            RegisterEvent(true, false, false, false, false);

        }

        /**
         * This method is called when a FullScreenChanged event occurs in the UI object
         */
        private void mediaPlayer_FullScreenChanged(object sender, EventArgs e)
        {
            _fullScreen = _myPlayer.IsFullScreen.ToString();

            // Do not report before init time has passed
            if (!_initDone)
            {
                return;
            }

            // Register that the current event is a FullScreenChanged
            RegisterEvent(true, false, false, false, true);
        }

        /**
         * Starts a thread that collects initial information
         */
        private void StartInitThread()
        {
            new Thread(DoInitThread).Start();
        }

        /**
         * Retrieves our IP and AS number. Also retrieves the avaliable video bitrates from the server
         * This method needs to finish before other information events will be sent.
         * This is done separately because it only needs to be done once
         */
        public void DoInitThread()
        {
            String ourIPinfo = "";
            String avaliableVideoRates = "";

            // While the wanted parameters is not returned, try again.
            while (String.IsNullOrWhiteSpace(ourIPinfo) || String.IsNullOrWhiteSpace(avaliableVideoRates))
            {

                // If IP and AS is not retrieved, retrieve it
                if (String.IsNullOrWhiteSpace(ourIPinfo))
                {
                    ourIPinfo = GetMyIpAndAS();
                }

                // If rates are not retireved, retrieve them
                if (String.IsNullOrWhiteSpace(avaliableVideoRates))
                {
                    avaliableVideoRates = GetAvaliableVideoBitrates();
                }

                // The video rates have a tendency to return 0 before they are set, check for it
                if (avaliableVideoRates.Equals("0"))
                {
                    avaliableVideoRates = "";
                }

                // Wait before trying again
                Thread.Sleep(100);
            }

            // We have our variables
            String[] ipInfoArray = ourIPinfo.Split(';');
            _myIp = ipInfoArray[0].Trim();
            _ispAS = ipInfoArray[1].Trim();
            _avaliableVideoBitrates = avaliableVideoRates;


            /**
             * Set _previousVideoBitrate to be what we are running at now.
             * Next videoBitrateChange will then be able to know if it is UP/DOWN
             */
            _previousVideoBitrate = _videoBitrate;

            // Register that the current event is an InitialInformation
            RegisterEvent(false, false, true, false, false);

            // Init is Done
            _initDone = true;
        }

        /**
         * This method starts a separate thread which gathers information and sends it
         */
        private void StartInfoThread()
        {
            new Thread(DoInfoThread).Start();
        }

        /**
         * The method for information gathering
         */
        public void DoInfoThread()
        {
            while (true)            
            {

                // Wait for signal
                autoEvent.WaitOne();

                // Empty queue
                while (eventQ.Count > 0)
                {

                    // Get a registered event
                    PostInformation data = eventQ.FirstOrDefault();

                    /* 
                    * Send all my information to the home gateway testbed and retrieve server response, print this to the TextBlock.
                    * We assume that the click22 will always translate to the correct router through custom DNS lookup tables.
                    */
                    SetDebugTextBlock(PostMyInformation("http://click22.item.ntnu.no:23500/receiveParams2.php", data));

                    // Send a copy of the information to a statistics server
                    PostMyInformation("http://cap.item.ntnu.no/logger/logScript.php", data);
               
                    // Then remove it, task done
                    eventQ.RemoveAt(0);
                }

            }

        }

        // This method starts a separate thread that registeres an event periodically
        private void StartPeriodicMeasurmentThread()
        {
            new Thread(DoPeriodicMeasurement).Start();
        }

        // This method is responsible for periodic notifications
        private void DoPeriodicMeasurement()
        {
            // Loop forever, send message every 1 min
            while (true)
            {

                // Check if the player is Paused or Stopped
                Boolean isPausedOrStopped = _latestPlayState.Equals("Paused") || _latestPlayState.Equals("Stopped");

                // If the trancient period is over, produce a periodic event as long as playstate is not Paused or Stopped
                if (_initDone && !isPausedOrStopped)
                {
                    RegisterEvent(false, false, false, true, false);
                }

                // Sleep one minute
                Thread.Sleep(60000);
            }

        }

        /**
         * Helper method to put events in queue for sending
         */
        private void RegisterEvent(bool playStateEvent, bool bitrateEvent, bool initialInfo, bool periodic, bool fullScreen)
        {

            // Has the video been paused or similar actions?
            if (playStateEvent)
            {
                _eventCause = "PlayStateChanged";
                _eventMessage = _latestPlayState;
            }

            // Has the bit rate changed?
            if (bitrateEvent)
            {
                _eventCause = "DownloadBitrateChanged";

                int previousRate = Int32.Parse(_previousVideoBitrate);
                int newRate = Int32.Parse(_videoBitrate);

                if (newRate > previousRate)
                {
                    _eventMessage = "UP";
                }
                else
                {
                    _eventMessage = "DOWN";
                }

                // Remember the newRate for next comparison
                _previousVideoBitrate = newRate.ToString();

            }

            // Is this the initial information package?
            if (initialInfo)
            {
                _eventCause = "InitialInformation";
                _eventMessage = "DONE";
            }

            // Or is it a periodic notification?
            if (periodic)
            {
                _eventCause = "PeriodicNotification";
                _eventMessage = "ALIVE";
            }

            if (fullScreen)
            {
                _eventCause = "FullScreenChanged";
                _eventMessage = _myPlayer.IsFullScreen.ToString();
            }

            /**
             * FINALLY
             * Allocate information for a PostInformation object and add it to the Queue
             */

            // Retrieve the timeIncrement and guid
            String timeIncrement = GetTimeSinceStart();
            String guid = _guid.ToString();

            /** 
             * We cannot retrieve this information here, it will halt the video player.
             * We need to do it in a separate thread.
             */
            String avgProcessorLoad = "";
            String avgProcessLoad = "";
            String sourceIpAndPort = "";

            // Put all info in a PostInformation object
            PostInformation informationObject = new PostInformation(_eventCause, _eventMessage, _latestPlayState, _fullScreen, guid, avgProcessorLoad, avgProcessLoad, _videoBitrate, _avaliableVideoBitrates, _myIp, _ispAS, sourceIpAndPort, timeIncrement);

            // Start a Task to retrieve the missing parameters and put this informationObject in the eventQ
            RegistrationTask task = new RegistrationTask(this, eventQ, informationObject);
            task.DoWork();

        }

        /**
         * Uses the system analytics object to retireve processor load
         */
        public String GetAverageProcessorLoad()
        {

            // init result string
            String result = "";

            // execute a query for avg cpu usage from the Dispatcher
            Dispatcher.BeginInvoke(delegate()
            {
                result = _analytics.AverageProcessorLoad.ToString();
            });

            // It is async, wait for reply
            while (String.IsNullOrWhiteSpace(result))
            {
                System.Threading.Thread.Sleep(5);
            }

            // Round the float to 3 digits
            float number = float.Parse(result);
            double numberRounded = Math.Round(number, 1);
            result = numberRounded.ToString();

            // then return
            return result;

        }

        /**
         * Uses the system analytics object to retireve Silverlight process load
         */
        public String GetAverageProcessLoad()
        {

            // init result string
            String result = "";

            // execute a query for this process avg cpu usage from the Dispatcher
            Dispatcher.BeginInvoke(delegate()
            {
                result = _analytics.AverageProcessLoad.ToString();
            });

            // It is async, wait for reply
            while (String.IsNullOrWhiteSpace(result))
            {
                System.Threading.Thread.Sleep(5);
            }

            // Round the float to 3 digits
            float number = float.Parse(result);
            double numberRounded = Math.Round(number, 1);
            result = numberRounded.ToString();

            // then return
            return result;

        }

        /**
         * Asks the media player for source address and port
         */
        public String GetSourceIpAndPort()
        {

            // Init result string
            string ip = "";
            string port = "";


            // use dispatcher for cross-thread calls
            _myPlayer.Dispatcher.BeginInvoke(delegate()
            {
                ip = _myPlayer.CurrentPlaylistItem.MediaSource.DnsSafeHost;
                port = _myPlayer.CurrentPlaylistItem.MediaSource.Port.ToString();
            });

            // It is async, wait for reply
            while (String.IsNullOrWhiteSpace(ip) || String.IsNullOrWhiteSpace(port))
            {
                System.Threading.Thread.Sleep(5);
            }

            // Return it
            return ip + ":" + port;
        }

        /**
         * Asks the media player for the avaliable bitrates
         */
        public String GetAvaliableVideoBitrates()
        {

            // init result string
            String result = "";

            // Init the result array and set the first value to -1, then we can determine if it has been changed
            long[] array = new long[1];
            array[0] = -1L;

            // use dispatcher for cross-thread calls
            _myPlayer.Dispatcher.BeginInvoke(delegate()
            {
                array = _myPlayer.AvailableVideoBitrates.ToArray();

            });

            // It is async, wait for reply
            while (String.IsNullOrWhiteSpace(result))
            {

                // Check if the first value differs from what we initialized. If it does, our array have been updated
                if (array.FirstOrDefault() != -1L)
                {
                    // First value is not -1L, we have retrieved a new array. Lets see if we have received an empty array
                    if (array.FirstOrDefault() != 0)
                    {
                        // It was not empty, add all information elements, separate by " - " until we reach the next to last element
                        for (int i = 0; i < array.Count() - 1; i++)
                        {
                            result += array[i].ToString() + " - ";
                        }

                        // Add the last element without " - " after
                        result += array[array.Count() - 1].ToString();
                    }
                    else
                    {
                        // We have received a empty or otherwise wrong array, add firstOrDefault value(often 0)
                        result = array.FirstOrDefault().ToString();
                    }
                }

                // The array[0] have not changed, wait 5 ms and check again
                System.Threading.Thread.Sleep(5);
            }

            // then return
            return result;

        }

        /**
         * Sends a message to a php script to obtain our IP address and AS information
         */
        public String GetMyIpAndAS()
        {

            SendHttpMessage("http://cap.item.ntnu.no/getIpAndAS.php", "get", "dummy");

            // It is asnyc, wait for response
            while (_httpDownResult.Equals(""))
            {
                System.Threading.Thread.Sleep(5);
            }

            // we have our response
            String ipResult = _httpDownResult;
            _httpDownResult = "";

            // Strip it for html tags
            ipResult = ipResult.Replace("<html>", "");
            ipResult = ipResult.Replace("</html>", "");
            ipResult = ipResult.Replace("<body>", "");
            ipResult = ipResult.Replace("</body>", "");
            ipResult = ipResult.Trim();

            // Update local variable routerIp if ipResult is not null, empty or whitespaces AND routerIp is not already correct
            if (!String.IsNullOrWhiteSpace(ipResult) && !ipResult.Equals(_myIp))
            {
                _myIp = ipResult;
            }

            // Return the IP we found
            return ipResult;
        }

        /**
         * Takes some parameters and reads local variables, creates a HTTP message and sends it to the router
         */
        public String PostMyInformation(String serverUrl, PostInformation postInformation)
        {
            _httpUpResult = "";

            // Set header content-type
            _myClient.Headers["Content-Type"] = "application/x-www-form-urlencoded";

            //and send the POST message
            SendHttpMessage(serverUrl, "post",
                "event=" + postInformation.eventName
                + "&eventMessage=" + postInformation.eventMessage
                + "&playState=" + postInformation.playState
                + "&fullScreen=" + postInformation.fullScreen
                + "&clientId=" + postInformation.clientId
                + "&avgProcessorLoad=" + postInformation.avgProcessorLoad
                + "&avgProcessLoad=" + postInformation.avgProcessLoad
                + "&videoDownloadBitrate=" + postInformation.videoDownloadBitrate
                + "&avaliableVideoBitrates=" + postInformation.avaliableVideoBitrates
                + "&myIp=" + postInformation.myIp
                + "&ispAS=" +postInformation.isp
                + "&sourceIpAndPort=" + postInformation.sourceIpAndPort
                + "&timeIncrement=" + postInformation.timeIncrement);

            // Call is async, wait for reply
            while (_httpUpResult.Equals(""))
            {
                System.Threading.Thread.Sleep(5);
            }

            // Reply has arrived, read it and set result to ""
            String postResult = _httpUpResult;
            _httpUpResult = "";

            // remove html tags from response
            postResult = postResult.Replace("<html>", "");
            postResult = postResult.Replace("</html>", "");
            postResult = postResult.Replace("<body>", "");
            postResult = postResult.Replace("</body>", "");
            postResult = postResult.Trim();

            // Return the result
            return postResult;
        }

        /**
        * This method returns the number of seconds since the application was started
        */
        private String GetTimeSinceStart()
        {
            // Calculate timeIncrement
            TimeSpan ts = (DateTime.UtcNow - _clientStartUpTime);
            double timeIncrement = ts.TotalSeconds;

            // Round to two decimals in number of seconds
            String roundedTimeIncrement = Math.Round(timeIncrement, 2).ToString();

            return roundedTimeIncrement;
        }

        /**
         * Helper method for the actual sending of HTTP messages
         */
        private void SendHttpMessage(String url, String type, String postMessage)
        {

            if (type.Equals("get"))
            {
                // get method
                _myClient.DownloadStringCompleted += new DownloadStringCompletedEventHandler(svcDownloadCompleted);
                _myClient.DownloadStringAsync(new Uri(url));
            }

            if (type.Equals("post"))
            {
                // post method
                _myClient.UploadStringCompleted += new UploadStringCompletedEventHandler(svcUploadCompleted);
                _myClient.UploadStringAsync(new Uri(url), "POST", postMessage);
            }
        }

        /**
         * Helper method for the actual sending of HTTP messages, will be called when the http action is done
         */
        void svcDownloadCompleted(object sender, DownloadStringCompletedEventArgs e)
        {
            if (e.Error == null)
            {
                _httpDownResult = e.Result;
            }
            else
            {
                _httpDownResult = e.Error.ToString();
            }

        }

        /**
         * Helper method for the actual sending of HTTP messages, will be called when the http action is done
         */
        void svcUploadCompleted(object sender, UploadStringCompletedEventArgs e)
        {
            if (e.Error == null)
            {
                _httpUpResult = e.Result.ToString();
            }
            else
            {
                _httpUpResult = e.Error.ToString();
            }
        }

        /**
         * For printing information to the textblock
         */
        public void SetDebugTextBlock(String toSet)
        {
            _infoBlock.Dispatcher.BeginInvoke(delegate()
            {
                _infoBlock.Text = toSet;
            });
        }

        /**
         * To pass parameters to a Thread, using the Task pattern
         */
        public class RegistrationTask
        {
            // class variables, taken as parameter
            private List<PostInformation> eventQ;
            private MainPage mainPage;
            private PostInformation informationObject;

            // Constructor, needs reference to mainPage, eventQ and informationObject
            public RegistrationTask(MainPage mainPage, List<PostInformation> eventQ, PostInformation informationObject)
            {
                this.mainPage = mainPage;
                this.eventQ = eventQ;
                this.informationObject = informationObject;
            }

            // The public method that starts the thread
            public void DoWork()
            {
                Thread t = new Thread(new ThreadStart(DoWorkCore));
                t.Start();
            }

            // This is the working method that performs the last information gathering
            private void DoWorkCore()
            {
                // Retrieve information
                String avgProcessorLoad = mainPage.GetAverageProcessorLoad();
                String avgProcessLoad = mainPage.GetAverageProcessLoad();
                String sourceIpAndPort = mainPage.GetSourceIpAndPort();

                // Set the info to the correct fields
                informationObject.avgProcessorLoad = avgProcessorLoad;
                informationObject.avgProcessLoad = avgProcessLoad;
                informationObject.sourceIpAndPort = sourceIpAndPort;

                /** 
                 * To protect the application from potential failiure we check if the queue is processed right.
                 * If the queue grows too large, some error has occured in the sending thread. 1000 in queue, is WAY too much
                 */
                if (eventQ.Count > 1000) { return; }

                // Add it to the queue
                eventQ.Add(informationObject);

                // Notify the processing thread
                autoEvent.Set();

            }
        }


        /**
         * This is an information object which holds a "snapshot" of all the variables at the time of an event occuring
         */
        public class PostInformation
        {

            // A shitload of variables with getters and setters
            public String eventName { get; set; }
            public String eventMessage { get; set; }
            public String playState { get; set; }
            public String fullScreen { get; set; }
            public String clientId { get; set; }
            public String avgProcessorLoad { get; set; }
            public String avgProcessLoad { get; set; }
            public String videoDownloadBitrate { get; set; }
            public String avaliableVideoBitrates { get; set; }
            public String myIp { get; set; }
            public String isp { get; set; }
            public String sourceIpAndPort { get; set; }
            public String timeIncrement { get; set; }


            // Constructor for a one-line instantiation of an object
            public PostInformation(String eventName, String eventMessage, String playState, String fullScreen, String clientId, String avgProcessorLoad, String avgProcessLoad, String videoDownloadBitrate, String avaliableVideoBitrates, String myIp, String isp, String sourceIpAndPort, String timeIncrement)
            {
                this.eventName = eventName;
                this.eventMessage = eventMessage;
                this.playState = playState;
                this.fullScreen = fullScreen;
                this.clientId = clientId;
                this.avgProcessorLoad = avgProcessorLoad;
                this.avgProcessLoad = avgProcessLoad;
                this.videoDownloadBitrate = videoDownloadBitrate;
                this.avaliableVideoBitrates = avaliableVideoBitrates;
                this.myIp = myIp;
                this.isp = isp;
                this.sourceIpAndPort = sourceIpAndPort;
                this.timeIncrement = timeIncrement;
            }
        }

    }
}